"""GLEIF API."""
from .api import GLEIFResponse
from .const import URL_API
from .utils import load_json


class PyGleif:
    """Query GLEIF API and return response."""

    def __init__(self, lei_code: str) -> None:
        """Init class."""
        json_data = load_json(search_url=URL_API, search_string=lei_code)
        self.response = GLEIFResponse(**json_data)



"""Data."""
from datetime import datetime
from typing import Any, List, Optional

from pydantic import BaseModel, Field


class ValidatedAt(BaseModel):
    """Represent validated at information."""

    id: str = Field(alias="id")
    other: Optional[str] = Field(alias="other")


class Registration(BaseModel):
    """Represent registration information."""

    corroboration_level: str = Field(alias="corroborationLevel")
    initial_registration_date: datetime = Field(alias="initialRegistrationDate")
    last_update_date: datetime = Field(alias="lastUpdateDate")
    managing_lou: str = Field(alias="managingLou")
    next_renewal_date: datetime = Field(alias="nextRenewalDate")
    other_validation_authorities: List[Any] = Field(alias="otherValidationAuthorities")
    status: str = Field(alias="status")
    validated_as: str = Field(alias="validatedAs")
    validated_at: ValidatedAt = Field(alias="validatedAt")


class GeneralEntity(BaseModel):
    """Represent a general entity ."""

    lei: Optional[str] = Field(alias="lei")
    name: Optional[str] = Field(alias="name")


class Address(BaseModel):
    """Represent address ."""

    language: str = Field(alias="language")
    address_lines: List[str] = Field(alias="addressLines")
    address_number: Optional[str] = Field(alias="addressNumber")
    address_number_within_building: Optional[str] = Field(
        alias="addressNumberWithinBuilding"
    )
    mail_routing: Optional[str] = Field(alias="mailRouting")
    city: str = Field(alias="city")
    region: Optional[str] = Field(alias="region")
    country: str = Field(alias="country")
    postal_code: Optional[str] = Field(alias="postalCode")


class Expiration(BaseModel):
    """Represent expiration data."""

    date: Optional[str] = Field(alias="date")
    reason: Optional[str] = Field(alias="reason")


class LegalForm(BaseModel):
    """Represent the legal form."""

    id: Optional[str] = Field(alias="id")
    other: Optional[str] = Field(alias="other")


class Name(BaseModel):
    """Represent the name."""

    language: str = Field(alias="language")
    name: str = Field(alias="name")
    type: Optional[str] = Field(alias="type")


class RegisteredAt(BaseModel):
    """Represent registered at."""

    id: str = Field(alias="id")
    other: Optional[str] = Field(alias="other")


class Entity(BaseModel):
    """Represent entity information."""

    associated_entity: GeneralEntity = Field(alias="associatedEntity")
    category: str = Field(alias="category")
    creation_date: Optional[str] = Field(alias="creationDate")
    event_groups: List[Any] = Field(alias="eventGroups")
    expiration: Expiration = Field(alias="expiration")
    headquarters_address: Address = Field(alias="headquartersAddress")
    jurisdiction: str = Field(alias="jurisdiction")
    legal_address: Address = Field(alias="legalAddress")
    legal_form: LegalForm = Field(alias="legalForm")
    legal_name: Name = Field(alias="legalName")
    other_addresses: List[Any] = Field(alias="otherAddresses")
    other_names: List[Name] = Field(alias="otherNames")
    registered_as: str = Field(alias="registeredAs")
    registered_at: RegisteredAt = Field(alias="registeredAt")
    status: str = Field(alias="status")
    successor_entities: List[Any] = Field(alias="successorEntities")
    sub_category: Optional[str] = Field(alias="subCategory")
    successor_entity: GeneralEntity = Field(alias="successorEntity")
    transliteraded_other_names: List[Any] = Field(alias="transliteratedOtherNames")


class Attributes(BaseModel):
    """Represent attribute information."""

    bic: Optional[List[str]] = Field(alias="bic")
    lei: str = Field(alias="lei")
    entity: Entity = Field(alias="entity")
    registration: Registration = Field(alias="registration")


class LinkData(BaseModel):
    """Represent a link ."""

    self: Optional[str] = Field(alias="self")
    related: Optional[str] = Field(alias="related")
    reporting_exception: Optional[str] = Field(alias="reporting-exception")


class Links(BaseModel):
    """Represent a link ."""

    links: LinkData = Field(alias="links")


class Relationships(BaseModel):
    """Represent a relationships ."""

    managing_lou: Links = Field(alias="managing-lou")
    lei_issuer: Links = Field(alias="lei-issuer")
    field_modifications: Links = Field(alias="field-modifications")
    direct_parent: Links = Field(alias="direct-parent")
    ultimate_parent: Links = Field(alias="ultimate-parent")


class Data(BaseModel):
    """Represent data information."""

    id: str = Field(alias="id")
    type: str = Field(alias="type")
    attributes: Attributes = Field(alias="attributes")
    links: LinkData = Field(alias="links")
    relationships: Relationships = Field(alias="relationships")



"""Constants."""
URL_API = "https://api.gleif.org/api/v1/lei-records/"
URL_SEARCH = "https://api.gleif.org/api/v1/lei-records?filter%5Bfulltext%5D="


# The ELF codes are unique per country, not globally.
# Hence the nested dict.
LEGAL_FORMS = {
    "NO": {
        "326Y": "Ansvarlig selskap",
        "3C7U": "Stiftelse",
        "3L58": "Kommunalt foretak",
        "4ZRR": "Norskregistrert utenlandsk foretak",
        "50TD": "Statsforetak",
        "5ZTZ": "Fylkeskommunalt foretak",
        "8S9H": "Gjensidig forsikringsselskap",
        "9DI1": "Boligbyggelag",
        "AEV1": "Partsrederi",
        "BJ65": "Pensjonskasse",
        "CF5L": "Ansvarlig selskap, delt ansvar",
        "DRPL": "Europeisk økonomisk foretaksgruppe",
        "EXD7": "Enkeltpersonforetak",
        "FSBD": "Forening/lag/innretning (eller bare forening?)",
        "GYY6": "Annen juridisk person",
        "IQGE": "Allmennaksjeselskap",
        "K5P8": "Samvirkeforetak",
        "LJJW": "Verdipapirfond",
        "M9IQ": "Kommandittselskap",
        "O0EU": "Borettslag",
        "O7LB": "Europeisk samvirkeforetak",
        "PB3V": "Konkursbo",
        "Q0Q1": "Selskap med begrenset ansvar",
        "R71C": "Sparebank",
        "V06W": "Europeisk selskap",
        "YI42": "Aksjeselskap",
        "YTMC": "Interkommunalt selskap",
        "ZQ0Q": "Eierseksjonssameie",
    },
    "SE": {
        "1TN0": "Ideell förening (som bedriver näringsverksamhet)",
        "2UAX": "Trossamfund (som bedriver näringsverksamhet)",
        "381R": "Medlemsbank",
        "54P7": "Europabolag",
        "9YIP": "Europakooperativ",
        "AZTO": "Sambruksförening",
        "BEAY": "Handelsbolag",
        "BYQJ": "Bankaktiebolag",
        "C61P": "Ekonomisk förening",
        "CX05": "Kommanditbolag",
        "G04R": "Sparbank",
        "M0Y0": "Försäkringsförening",
        "O1QI": "Ömsesidigt försäkringsbolag",
        "OJ9I": "Försäkringsaktiebolag",
        "PDQ0": "Kooperativ hyresrättsförening",
        "RLJO": "Europeisk ekonomisk intressegruppering",
        "RYFP": "Europeisk gruppering för territoriellt samarbete",
        "SSOM": "Bostadsrättsförening",
        "WZDB": "Bostadsförening",
        "XJHM": "Aktiebolag",
    },
}



"""Utility functions."""
from __future__ import annotations

import json
from typing import Any, Dict, List, Union, cast
from urllib import request


def load_json(search_url: str, search_string: str) -> list[Any] | dict[Any, Any]:
    """Download data as JSON."""
    with request.urlopen(f"{search_url}{request.quote(search_string)}") as fdesc:
        return cast(Union[Dict[Any, Any], List[Any]], json.loads(fdesc.read()))



"""Search."""
from __future__ import annotations

from typing import Optional

from pygleif.api import SearchResponse
from pygleif.const import URL_SEARCH

from .utils import load_json


class Search:
    """Class to use the search form of the GLEIF web site."""

    response = Optional[SearchResponse]

    def __init__(self, orgnr: str) -> None:
        """Init class."""
        json_data = load_json(search_url=URL_SEARCH, search_string=orgnr)

        if json_data["data"]:
            self.response = SearchResponse(**json_data)
        else:
            self.response = None



"""Meta data."""
from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field


class GoldenCopy(BaseModel):
    """Represent golden copy information."""

    publish_date: datetime = Field(alias="publishDate")


class Pagination(BaseModel):
    """Represent response pagination."""

    current_page: int = Field(alias="currentPage")
    per_page: int = Field(alias="perPage")
    _from: int = Field(alias="from")
    to: int = Field(alias="to")
    total: int = Field(alias="total")
    last_page: int = Field(alias="lastPage")


class Meta(BaseModel):
    """Represent meta information."""

    golden_copy: GoldenCopy = Field(alias="goldenCopy")
    # Pagination is part of the search response
    pagination: Optional[Pagination] = Field(alias="pagination")



"""Pydantic Representation of models."""
from typing import List

from pydantic import BaseModel, Field

from .data import Data
from .meta import Meta


class GLEIFResponse(BaseModel):
    """Represent a base response."""

    meta: Meta = Field(alias="meta")
    data: Data = Field(alias="data")


class SearchResponse(BaseModel):
    """Represent search result response."""

    meta: Meta = Field(alias="meta")
    data: List[Data] = Field(alias="data")
